@@ -1,12 +1,8 @@
-#This file is part of Tryton.  The COPYRIGHT file at the top level of
-#this repository contains the full copyright notices and license terms.
-__all__ = ['PYSONEncoder', 'PYSONDecoder', 'Eval', 'Not', 'Bool', 'And', 'Or',
-    'Equal', 'Greater', 'Less', 'If', 'Get', 'In', 'Date', 'DateTime', 'Len']
-try:
-    import simplejson as json
-except ImportError:
-    import json
+# This file is part of Tryton.  The COPYRIGHT file at the top level of
+# this repository contains the full copyright notices and license terms.
+import json
 import datetime
+from decimal import Decimal
 from dateutil.relativedelta import relativedelta
 from functools import reduce
 
@@ -24,36 +20,40 @@ class PYSON(object):
         raise NotImplementedError
 
     def __invert__(self):
-        if self.types() != set([bool]):
+        if self.types() != {bool}:
             return Not(Bool(self))
         else:
             return Not(self)
 
     def __and__(self, other):
+        if (isinstance(other, PYSON)
+                and other.types() != {bool}):
+            other = Bool(other)
         if (isinstance(self, And)
                 and not isinstance(self, Or)):
             self._statements.append(other)
             return self
-        if (isinstance(other, PYSON)
-                and other.types() != set([bool])):
-            other = Bool(other)
-        if self.types() != set([bool]):
+        if self.types() != {bool}:
             return And(Bool(self), other)
         else:
             return And(self, other)
 
+    __rand__ = __and__
+
     def __or__(self, other):
+        if (isinstance(other, PYSON)
+                and other.types() != {bool}):
+            other = Bool(other)
         if isinstance(self, Or):
             self._statements.append(other)
             return self
-        if (isinstance(other, PYSON)
-                and other.types() != set([bool])):
-            other = Bool(other)
-        if self.types() != set([bool]):
+        if self.types() != {bool}:
             return Or(Bool(self), other)
         else:
             return Or(self, other)
 
+    __ror__ = __or__
+
     def __eq__(self, other):
         return Equal(self, other)
 
@@ -81,6 +81,14 @@ class PYSON(object):
     def contains(self, k):
         return In(k, self)
 
+    def __repr__(self):
+        klass = self.__class__.__name__
+        return '%s(%s)' % (klass, ', '.join(map(repr, self.__repr_params__)))
+
+    @property
+    def __repr_params__(self):
+        return NotImplementedError
+
 
 class PYSONEncoder(json.JSONEncoder):
 
@@ -94,29 +102,41 @@ class PYSONEncoder(json.JSONEncoder):
                         ).pyson()
             else:
                 return Date(obj.year, obj.month, obj.day).pyson()
+        elif isinstance(obj, Decimal):
+            return float(obj)
         return super(PYSONEncoder, self).default(obj)
 
 
 class PYSONDecoder(json.JSONDecoder):
 
-    def __init__(self, context=None):
+    def __init__(self, context=None, noeval=False):
         self.__context = context or {}
+        self.noeval = noeval
         super(PYSONDecoder, self).__init__(object_hook=self._object_hook)
 
     def _object_hook(self, dct):
         if '__class__' in dct:
-            klass = globals().get(dct['__class__'])
-            if klass and hasattr(klass, 'eval'):
-                return klass.eval(dct, self.__context)
+            klass = CONTEXT.get(dct['__class__'])
+            if klass:
+                if not self.noeval:
+                    return klass.eval(dct, self.__context)
+                else:
+                    dct = dct.copy()
+                    del dct['__class__']
+                    return klass(**dct)
         return dct
 
 
 class Eval(PYSON):
 
-    def __init__(self, value, default=''):
+    def __init__(self, v, d=''):
         super(Eval, self).__init__()
-        self._value = value
-        self._default = default
+        self._value = v
+        self._default = d
+
+    @property
+    def __repr_params__(self):
+        return self._value, self._default
 
     def pyson(self):
         return {
@@ -129,7 +149,7 @@ class Eval(PYSON):
         if isinstance(self._default, PYSON):
             return self._default.types()
         else:
-            return set([type(self._default)])
+            return {type(self._default)}
 
     @staticmethod
     def eval(dct, context):
@@ -138,13 +158,18 @@ class Eval(PYSON):
 
 class Not(PYSON):
 
-    def __init__(self, value):
+    def __init__(self, v):
         super(Not, self).__init__()
-        if isinstance(value, PYSON):
-            assert value.types() == set([bool]), 'value must be boolean'
-        else:
-            assert isinstance(value, bool), 'value must be boolean'
-        self._value = value
+        if isinstance(v, PYSON):
+            if v.types() != {bool}:
+                v = Bool(v)
+        elif not isinstance(v, bool):
+            v = bool(v)
+        self._value = v
+
+    @property
+    def __repr_params__(self):
+        return (self._value,)
 
     def pyson(self):
         return {
@@ -153,7 +178,7 @@ class Not(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -162,9 +187,13 @@ class Not(PYSON):
 
 class Bool(PYSON):
 
-    def __init__(self, value):
+    def __init__(self, v):
         super(Bool, self).__init__()
-        self._value = value
+        self._value = v
+
+    @property
+    def __repr_params__(self):
+        return (self._value,)
 
     def pyson(self):
         return {
@@ -173,7 +202,7 @@ class Bool(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -182,17 +211,21 @@ class Bool(PYSON):
 
 class And(PYSON):
 
-    def __init__(self, *statements):
+    def __init__(self, *statements, **kwargs):
         super(And, self).__init__()
-        for statement in statements:
+        statements = list(statements) + kwargs.get('s', [])
+        for i, statement in enumerate(list(statements)):
             if isinstance(statement, PYSON):
-                assert statement.types() == set([bool]), \
-                    'statement must be boolean'
-            else:
-                assert isinstance(statement, bool), \
-                    'statement must be boolean'
+                if statement.types() != {bool}:
+                    statements[i] = Bool(statement)
+            elif not isinstance(statement, bool):
+                statements[i] = bool(statement)
         assert len(statements) >= 2, 'must have at least 2 statements'
-        self._statements = list(statements)
+        self._statements = statements
+
+    @property
+    def __repr_params__(self):
+        return tuple(self._statements)
 
     def pyson(self):
         return {
@@ -201,7 +234,7 @@ class And(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -222,20 +255,25 @@ class Or(And):
 
 class Equal(PYSON):
 
-    def __init__(self, statement1, statement2):
+    def __init__(self, s1, s2):
+        statement1, statement2 = s1, s2
         super(Equal, self).__init__()
         if isinstance(statement1, PYSON):
             types1 = statement1.types()
         else:
-            types1 = set([type(statement1)])
+            types1 = {type(s1)}
         if isinstance(statement2, PYSON):
             types2 = statement2.types()
         else:
-            types2 = set([type(statement2)])
+            types2 = {type(s2)}
         assert types1 == types2, 'statements must have the same type'
         self._statement1 = statement1
         self._statement2 = statement2
 
+    @property
+    def __repr_params__(self):
+        return (self._statement1, self._statement2)
+
     def pyson(self):
         return {
             '__class__': 'Equal',
@@ -244,7 +282,7 @@ class Equal(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
@@ -253,23 +291,29 @@ class Equal(PYSON):
 
 class Greater(PYSON):
 
-    def __init__(self, statement1, statement2, equal=False):
+    def __init__(self, s1, s2, e=False):
+        statement1, statement2, equal = s1, s2, e
         super(Greater, self).__init__()
         for i in (statement1, statement2):
             if isinstance(i, PYSON):
-                assert i.types().issubset(set([int, long, float])), \
+                assert i.types().issubset({int, float, type(None)}), \
                     'statement must be an integer or a float'
             else:
-                assert isinstance(i, (int, long, float)), \
+                assert isinstance(i, (int, float, type(None))), \
                     'statement must be an integer or a float'
         if isinstance(equal, PYSON):
-            assert equal.types() == set([bool])
-        else:
-            assert isinstance(equal, bool)
+            if equal.types() != {bool}:
+                equal = Bool(equal)
+        elif not isinstance(equal, bool):
+            equal = bool(equal)
         self._statement1 = statement1
         self._statement2 = statement2
         self._equal = equal
 
+    @property
+    def __repr_params__(self):
+        return (self._statement1, self._statement2, self._equal)
+
     def pyson(self):
         return {
             '__class__': 'Greater',
@@ -279,10 +323,21 @@ class Greater(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
+
+    @staticmethod
+    def _convert(dct):
+        for i in ('s1', 's2'):
+            if dct[i] is None:
+                dct[i] = 0.0
+            if not isinstance(dct[i], (int, float)):
+                dct = dct.copy()
+                dct[i] = float(dct[i])
+        return dct
 
     @staticmethod
     def eval(dct, context):
+        dct = Greater._convert(dct)
         if dct['e']:
             return dct['s1'] >= dct['s2']
         else:
@@ -298,6 +353,7 @@ class Less(Greater):
 
     @staticmethod
     def eval(dct, context):
+        dct = Less._convert(dct)
         if dct['e']:
             return dct['s1'] <= dct['s2']
         else:
@@ -306,27 +362,32 @@ class Less(Greater):
 
 class If(PYSON):
 
-    def __init__(self, condition, then_statement, else_statement=None):
+    def __init__(self, c, t, e=None):
+        condition, then_statement, else_statement = c, t, e
         super(If, self).__init__()
         if isinstance(condition, PYSON):
-            assert condition.types() == set([bool]), \
-                'condition must be boolean'
-        else:
-            assert isinstance(condition, bool), 'condition must be boolean'
+            if condition.types() != {bool}:
+                condition = Bool(condition)
+        elif not isinstance(condition, bool):
+            condition = bool(condition)
         if isinstance(then_statement, PYSON):
             then_types = then_statement.types()
         else:
-            then_types = set([type(then_statement)])
+            then_types = {type(then_statement)}
         if isinstance(else_statement, PYSON):
-            assert then_types == else_statement.types(), \
-                'then and else statements must be the same type'
+            else_types = else_statement.types()
         else:
-            assert then_types == set([type(else_statement)]), \
-                'then and else statements must be the same type'
+            else_types = {type(else_statement)}
+        assert then_types == else_types, \
+            'then and else statements must be the same type'
         self._condition = condition
         self._then_statement = then_statement
         self._else_statement = else_statement
 
+    @property
+    def __repr_params__(self):
+        return (self._condition, self._then_statement, self._else_statement)
+
     def pyson(self):
         return {
             '__class__': 'If',
@@ -339,7 +400,7 @@ class If(PYSON):
         if isinstance(self._then_statement, PYSON):
             return self._then_statement.types()
         else:
-            return set([type(self._then_statement)])
+            return {type(self._then_statement)}
 
     @staticmethod
     def eval(dct, context):
@@ -351,20 +412,25 @@ class If(PYSON):
 
 class Get(PYSON):
 
-    def __init__(self, obj, key, default=''):
+    def __init__(self, v, k, d=''):
+        obj, key, default = v, k, d
         super(Get, self).__init__()
         if isinstance(obj, PYSON):
-            assert obj.types() == set([dict]), 'obj must be a dict'
+            assert obj.types() == {dict}, 'obj must be a dict'
         else:
             assert isinstance(obj, dict), 'obj must be a dict'
         self._obj = obj
         if isinstance(key, PYSON):
-            assert key.types() == set([str]), 'key must be a string'
+            assert key.types() == {str}, 'key must be a string'
         else:
-            assert type(key) == str, 'key must be a string'
+            assert isinstance(key, str), 'key must be a string'
         self._key = key
         self._default = default
 
+    @property
+    def __repr_params__(self):
+        return (self._obj, self._key, self._default)
+
     def pyson(self):
         return {
             '__class__': 'Get',
@@ -377,7 +443,7 @@ class Get(PYSON):
         if isinstance(self._default, PYSON):
             return self._default.types()
         else:
-            return set([type(self._default)])
+            return {type(self._default)}
 
     @staticmethod
     def eval(dct, context):
@@ -386,26 +452,31 @@ class Get(PYSON):
 
 class In(PYSON):
 
-    def __init__(self, key, obj):
+    def __init__(self, k, v):
+        key, obj = k, v
         super(In, self).__init__()
         if isinstance(key, PYSON):
-            assert key.types().issubset(set([str, int, long])), \
+            assert key.types().issubset({str, int}), \
                 'key must be a string or an integer or a long'
         else:
-            assert type(key) in [str, int, long], \
+            assert isinstance(key, (str, int)), \
                 'key must be a string or an integer or a long'
         if isinstance(obj, PYSON):
-            assert obj.types().issubset(set([dict, list])), \
+            assert obj.types().issubset({dict, list}), \
                 'obj must be a dict or a list'
-            if obj.types() == set([dict]):
-                assert type(key) == str, 'key must be a string'
+            if obj.types() == {dict}:
+                assert isinstance(key, str), 'key must be a string'
         else:
-            assert type(obj) in [dict, list]
-            if type(obj) == dict:
-                assert type(key) == str, 'key must be a string'
+            assert isinstance(obj, (dict, list))
+            if isinstance(obj, dict):
+                assert isinstance(key, str), 'key must be a string'
         self._key = key
         self._obj = obj
 
+    @property
+    def __repr_params__(self):
+        return (self._key, self._obj)
+
     def pyson(self):
         return {
             '__class__': 'In',
@@ -414,24 +485,33 @@ class In(PYSON):
             }
 
     def types(self):
-        return set([bool])
+        return {bool}
 
     @staticmethod
     def eval(dct, context):
-        return dct['k'] in dct['v']
+        if dct['v']:
+            return dct['k'] in dct['v']
+        else:
+            return False
 
 
 class Date(PYSON):
 
     def __init__(self, year=None, month=None, day=None,
-            delta_years=0, delta_months=0, delta_days=0):
+            delta_years=0, delta_months=0, delta_days=0, **kwargs):
+        year = kwargs.get('y', year)
+        month = kwargs.get('M', month)
+        day = kwargs.get('d', day)
+        delta_years = kwargs.get('dy', delta_years)
+        delta_months = kwargs.get('dM', delta_months)
+        delta_days = kwargs.get('dd', delta_days)
         super(Date, self).__init__()
         for i in (year, month, day, delta_years, delta_months, delta_days):
             if isinstance(i, PYSON):
-                assert i.types().issubset(set([int, long, type(None)])), \
+                assert i.types().issubset({int, type(None)}), \
                     '%s must be an integer or None' % (i,)
             else:
-                assert isinstance(i, (int, long, type(None))), \
+                assert isinstance(i, (int, type(None))), \
                     '%s must be an integer or None' % (i,)
         self._year = year
         self._month = month
@@ -440,6 +520,11 @@ class Date(PYSON):
         self._delta_months = delta_months
         self._delta_days = delta_days
 
+    @property
+    def __repr_params__(self):
+        return (self._year, self._month, self._day,
+            self._delta_years, self._delta_months, self._delta_days)
+
     def pyson(self):
         return {
             '__class__': 'Date',
@@ -452,7 +537,7 @@ class Date(PYSON):
             }
 
     def types(self):
-        return set([datetime.date])
+        return {datetime.date}
 
     @staticmethod
     def eval(dct, context):
@@ -472,17 +557,25 @@ class DateTime(Date):
             hour=None, minute=None, second=None, microsecond=None,
             delta_years=0, delta_months=0, delta_days=0,
             delta_hours=0, delta_minutes=0, delta_seconds=0,
-            delta_microseconds=0):
+            delta_microseconds=0, **kwargs):
+        hour = kwargs.get('h', hour)
+        minute = kwargs.get('m', minute)
+        second = kwargs.get('s', second)
+        microsecond = kwargs.get('ms', microsecond)
+        delta_hours = kwargs.get('dh', delta_hours)
+        delta_minutes = kwargs.get('dm', delta_minutes)
+        delta_seconds = kwargs.get('ds', delta_seconds)
+        delta_microseconds = kwargs.get('dms', delta_microseconds)
         super(DateTime, self).__init__(year=year, month=month, day=day,
                 delta_years=delta_years, delta_months=delta_months,
-                delta_days=delta_days)
+                delta_days=delta_days, **kwargs)
         for i in (hour, minute, second, microsecond,
                 delta_hours, delta_minutes, delta_seconds, delta_microseconds):
             if isinstance(i, PYSON):
-                assert i.types() == set([int, long, type(None)]), \
+                assert i.types() == {int, type(None)}, \
                     '%s must be an integer or None' % (i,)
             else:
-                assert isinstance(i, (int, long, type(None))), \
+                assert isinstance(i, (int, type(None))), \
                     '%s must be an integer or None' % (i,)
         self._hour = hour
         self._minute = minute
@@ -493,6 +586,15 @@ class DateTime(Date):
         self._delta_seconds = delta_seconds
         self._delta_microseconds = delta_microseconds
 
+    @property
+    def __repr_params__(self):
+        date_params = super(DateTime, self).__repr_params__
+        return (date_params[:3]
+            + (self._hour, self._minute, self._second, self._microsecond)
+            + date_params[3:]
+            + (self._delta_hours, self._delta_minutes, self._delta_seconds,
+                self._delta_microseconds))
+
     def pyson(self):
         res = super(DateTime, self).pyson()
         res['__class__'] = 'DateTime'
@@ -507,11 +609,11 @@ class DateTime(Date):
         return res
 
     def types(self):
-        return set([datetime.datetime])
+        return {datetime.datetime}
 
     @staticmethod
     def eval(dct, context):
-        return datetime.datetime.now() + relativedelta(
+        return datetime.datetime.utcnow() + relativedelta(
             year=dct['y'],
             month=dct['M'],
             day=dct['d'],
@@ -529,17 +631,59 @@ class DateTime(Date):
             )
 
 
+class TimeDelta(PYSON):
+
+    def __init__(self, days=0, seconds=0, microseconds=0):
+        for i in [days, seconds, microseconds]:
+            if isinstance(i, PYSON):
+                assert i.types().issubset({int, float}), \
+                        '%s must be an integer' % (i,)
+            else:
+                assert isinstance(i, (int, float)), \
+                        '%s must be an integer' % (i,)
+        self._days = days
+        self._seconds = seconds
+        self._microseconds = microseconds
+
+    @property
+    def __repr_params__(self):
+        return self._days, self._seconds, self._microseconds
+
+    def pyson(self):
+        return {
+            '__class__': 'TimeDelta',
+            'd': self._days,
+            's': self._seconds,
+            'm': self._microseconds,
+            }
+
+    def types(self):
+        return {datetime.timedelta}
+
+    @staticmethod
+    def eval(dct, context):
+        return datetime.timedelta(
+            days=dct['d'],
+            seconds=dct['s'],
+            microseconds=dct['m'],
+            )
+
+
 class Len(PYSON):
 
-    def __init__(self, value):
+    def __init__(self, v):
         super(Len, self).__init__()
-        if isinstance(value, PYSON):
-            assert value.types().issubset(set([dict, list, str])), \
+        if isinstance(v, PYSON):
+            assert v.types().issubset({dict, list, str}), \
                 'value must be a dict or a list or a string'
         else:
-            assert type(value) in [dict, list, str], \
+            assert isinstance(v, (dict, list, str)), \
                 'value must be a dict or list or a string'
-        self._value = value
+        self._value = v
+
+    @property
+    def __repr_params__(self):
+        return (self._value,)
 
     def pyson(self):
         return {
@@ -548,7 +692,7 @@ class Len(PYSON):
             }
 
     def types(self):
-        return set([int, long])
+        return {int}
 
     @staticmethod
     def eval(dct, context):
@@ -569,5 +713,6 @@ CONTEXT = {
     'In': In,
     'Date': Date,
     'DateTime': DateTime,
+    'TimeDelta': TimeDelta,
     'Len': Len,
 }
